@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2020 Stefan Krah. All rights reserved.
+ * Copyright (c) 2020-2024 Stefan Krah. All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -7,12 +7,11 @@
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
- *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
  *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
@@ -32,6 +31,7 @@
 #include <sstream>
 #include <stdexcept>
 
+#include "mpdecimal.h"
 #include "decimal.hh"
 
 
@@ -181,14 +181,14 @@ Context& getcontext() {
 thread_local Context context{context_template};
 #endif
 
-/* Factory function for creating a context for maximum unrounded arithmetic.*/
+/* Factory function for creating a context for maximum unrounded arithmetic. */
 Context
 MaxContext()
 {
     return Context(maxcontext);
 }
 
-/* Factory function for creating IEEE interchange format contexts.*/
+/* Factory function for creating IEEE interchange format contexts. */
 Context
 IEEEContext(int bits)
 {
@@ -288,6 +288,10 @@ Decimal::ln10(int64_t n, Context& c)
   Decimal result;
   uint32_t status = 0;
 
+  if (n < 1 || n > MPD_MAX_PREC) {
+    throw ValueError("Decimal::ln10: prec argument must in [1, MAX_PREC]");
+  }
+
   mpd_ssize_t nn = util::safe_downcast<mpd_ssize_t, int64_t>(n);
   mpd_qln10(result.get(), nn, &status);
 
